import mock
import pytest
import responses
from responses import matchers

from gantry.exceptions import GantryException
from gantry.query.core.distance import GantryDistance

from ..conftest import ORIGIN, to_tz_time


@pytest.fixture
def gantry_distance_obj(api_client_obj):
    return GantryDistance(api_client_obj)


def get_query_content(series_obj_1, series_obj_2, diff_stat):
    return {
        "start_time": to_tz_time(series_obj_1.query_info.start_time),
        "end_time": to_tz_time(series_obj_1.query_info.end_time),
        "version": series_obj_1.query_info.version,
        "queries": {
            "query": {
                "query_type": "diff",
                "stat": diff_stat,
                "base": {
                    "query_type": "feature",
                    "feature": series_obj_1.name,
                    "model_node_id": series_obj_1.query_info.application_node_id,
                    "start_time": to_tz_time(series_obj_1.query_info.start_time),
                    "end_time": to_tz_time(series_obj_1.query_info.end_time),
                },
                "other": {
                    "query_type": "feature",
                    "feature": series_obj_2.name,
                    "model_node_id": series_obj_2.query_info.application_node_id,
                    "start_time": to_tz_time(series_obj_2.query_info.start_time),
                    "end_time": to_tz_time(series_obj_2.query_info.end_time),
                },
            }
        },
    }


@pytest.mark.parametrize("diff_stat", ["d1", "dinf", "ks", "kl"])
def test_diff_query(gantry_distance_obj, series_obj_1, series_obj_2, diff_stat):
    content = get_query_content(series_obj_1, series_obj_2, diff_stat)

    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/aggregate/query".format(ORIGIN),
            match=[matchers.json_params_matcher(content)],
            json={"response": "ok", "data": {"query": {diff_stat: 1500}}},
        )

        assert gantry_distance_obj._diff_query(series_obj_1, series_obj_2, diff_stat) == 1500


@pytest.mark.parametrize("diff_stat", ["d1", "dinf", "ks", "kl"])
def test_diff_query_invalid_api_response(
    gantry_distance_obj, series_obj_1, series_obj_2, diff_stat
):
    content = get_query_content(series_obj_1, series_obj_2, diff_stat)

    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/aggregate/query".format(ORIGIN),
            match=[matchers.json_params_matcher(content)],
            json={"response": "ok", "data": {"query": {diff_stat + "wrong_key": 1500}}},
        )

        with pytest.raises(RuntimeError):
            gantry_distance_obj._diff_query(series_obj_1, series_obj_2, diff_stat)


@pytest.mark.parametrize("diff_stat", ["d1", "dinf", "ks", "kl"])
def test_diff_query_error(gantry_distance_obj, series_obj_1, series_obj_2, diff_stat):
    with responses.RequestsMock() as rsps:
        rsps.add(
            responses.POST,
            "{}/api/v1/aggregate/query".format(ORIGIN),
            json={
                "response": "not_ok",
            },
            status=400,
        )

        with pytest.raises(GantryException):
            gantry_distance_obj._diff_query(series_obj_1, series_obj_2, diff_stat)


@mock.patch("gantry.query.core.distance.GantryDistance._diff_query")
def test_d1(mock_query, gantry_distance_obj, series_obj_1, series_obj_2):
    mock_query.return_value = 0.05
    assert gantry_distance_obj.d1(series_obj_1, series_obj_2) == 0.05
    mock_query.assert_called_once_with(series_obj_1, series_obj_2, "d1")


@mock.patch("gantry.query.core.distance.GantryDistance._diff_query")
def test_dinf(mock_query, gantry_distance_obj, series_obj_1, series_obj_2):
    mock_query.return_value = 0.45
    assert gantry_distance_obj.dinf(series_obj_1, series_obj_2) == 0.45
    mock_query.assert_called_once_with(series_obj_1, series_obj_2, "dinf")


@mock.patch("gantry.query.core.distance.GantryDistance._diff_query")
def test_ks(mock_query, gantry_distance_obj, series_obj_1, series_obj_2):
    mock_query.return_value = 1e-4
    assert gantry_distance_obj.ks(series_obj_1, series_obj_2) == 1e-4
    mock_query.assert_called_once_with(series_obj_1, series_obj_2, "ks")


@mock.patch("gantry.query.core.distance.GantryDistance._diff_query")
def test_kl(mock_query, gantry_distance_obj, series_obj_1, series_obj_2):
    mock_query.return_value = 100
    assert gantry_distance_obj.kl(series_obj_1, series_obj_2) == 100
    mock_query.assert_called_once_with(series_obj_1, series_obj_2, "kl")



import mock
import pytest

from gantry.automations.curators import Curator, UniformCurator, globals, main
from gantry.exceptions import ClientNotInitialized


@pytest.mark.parametrize(
    ["method", "params"],
    [
        ("get_all_curators", {}),
        ("get_curator", {"name": "test-curator"}),
        ("list_curators", {}),
    ],
)
def test_uninit_client_main(method, params):
    """Test all public methods from curator module
    fail if client is not initialized
    """
    with mock.patch("gantry.automations.curators.globals._CURATOR_CLIENT", None):
        with pytest.raises(ClientNotInitialized):
            getattr(main, method)(**params)


@pytest.mark.parametrize(
    ["curator_type", "params"],
    [
        (Curator, {"name": "test_curator", "application_name": "test-app"}),
        (UniformCurator, {"name": "test_curator", "application_name": "test-app", "limit": 10}),
    ],
)
def test_unint_curator(curator_type, params):
    """Test that curator initialization fails if client is not initialized"""
    with mock.patch("gantry.automations.curators.globals._API_CLIENT", None):
        with pytest.raises(ClientNotInitialized):
            curator_type(**params)


@pytest.mark.parametrize("backend", ["", "/some/file", "tcp://other/protocol"])
def test_global_init_error(backend):
    with mock.patch.dict("os.environ", {"GANTRY_LOGS_LOCATION": backend}):
        with pytest.raises(ValueError):
            main._init()


def test_no_api_key_provided():
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": ""}):
        with pytest.raises(ValueError):
            main._init(api_key=None)


def test_init():
    """Test init"""
    passed_api_key = "some_key"
    with mock.patch.dict(
        "os.environ",
        {
            "GANTRY_API_KEY": passed_api_key,
            "GANTRY_LOGS_LOCATION": "https://foo/bar",
        },
    ):
        main._init()
        assert globals._CURATOR_CLIENT._api_client._host == "https://foo/bar"
        assert globals._CURATOR_CLIENT._api_client._api_key == passed_api_key
        assert globals._API_CLIENT._host == "https://foo/bar"
        assert globals._API_CLIENT._api_key == passed_api_key


def test_init_default():
    """Test init client with default values"""
    main._init(api_key="ABCD1234")
    assert globals._CURATOR_CLIENT._api_client._host == "https://app.gantry.io"
    assert globals._CURATOR_CLIENT._api_client._api_key == "ABCD1234"
    assert globals._API_CLIENT._host == "https://app.gantry.io"
    assert globals._API_CLIENT._api_key == "ABCD1234"


@pytest.mark.parametrize(
    ["method", "params"],
    [
        ("get_all_curators", {}),
        ("get_curator", {"name": "test-curator"}),
        ("list_curators", {}),
    ],
)
def test_curator_methods(method, params):
    """Test all public methods from gantry module or gantry.main module
    resolve in the global _CURATOR_CLIENT methods
    """
    m = mock.Mock()
    with mock.patch("gantry.automations.curators.globals._CURATOR_CLIENT", m):
        getattr(main, method)(**params)
        getattr(m, method).assert_called_once_with(**params)



import inspect
from typing import Optional

from gantry.alerts.client import GantryAlerts
from gantry.exceptions import GantryException

# Singleton instance for alerts functionality within the scope of some backend and api key
_Alerts: Optional[GantryAlerts] = None


def _alerts_alias(func):
    """Decorator for GantryAlerts functions, exposed in main.py"""
    doc: str = ""
    signature = inspect.signature(func)
    orig_doc: Optional[str] = None
    if hasattr(GantryAlerts, func.__name__):
        doc = "Alias for :meth:`gantry.alerts.client.GantryAlerts.{}`".format(func.__name__)
        orig_doc = inspect.getdoc(getattr(GantryAlerts, func.__name__))
        signature = inspect.signature(getattr(GantryAlerts, func.__name__))

    if orig_doc:
        doc += "\n\n{}".format(orig_doc)
    signature = signature.replace(parameters=tuple(signature.parameters.values())[1:])
    func.__doc__ = doc
    func.__signature__ = signature

    return func


def validate_init():
    if _Alerts is None:
        raise GantryException("Alerts functionality has not been initialized.")



import uuid

import mock
import responses
from click.testing import CliRunner
from responses import matchers

from gantry.cli.bucket import create, list, update_secret

BUCKET_NAME = "test-bucket"
REGION = "us-west-2"
SECRET_NAME = "test_secret"

BUCKET_INFO = {
    "id": str(uuid.uuid4()),
    "bucket_name": BUCKET_NAME,
    "region": REGION,
    "storage_type": "S3",
    "organization_id": str(uuid.uuid4()),
    "secret_id": str(uuid.uuid4()),
    "notification_enabled": False,
    "created_at": "2022-12-15T00:05:23",
}


UNITTEST_HOST = "http://unittest"


@mock.patch("gantry.cli.bucket.APIClient.request")
def test_register_bucket_failure(mock_api_client):
    mock_api_client.return_value = {"response": "error", "error": "some error"}

    runner = CliRunner()
    cli_args = [
        "--name",
        BUCKET_NAME,
        "--region",
        REGION,
        "--storage-type",
        "s3",
        "--secret",
        SECRET_NAME,
    ]
    result = runner.invoke(create, cli_args, env={"GANTRY_API_KEY": "test"})
    assert result.exit_code == 1
    assert f"Fail to register bucket {BUCKET_NAME}: some error" in result.output


def test_register_bucket():
    runner = CliRunner()
    cli_args = [
        "--name",
        BUCKET_NAME,
        "--region",
        REGION,
        "--storage-type",
        "s3",
        "--secret",
        SECRET_NAME,
    ]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{UNITTEST_HOST}/api/v1/customer-resources/buckets",
            json={
                "response": "ok",
                "data": BUCKET_INFO,
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "bucket_name": BUCKET_NAME,
                        "region": REGION,
                        "storage_type": "S3",
                        "secret_name": SECRET_NAME,
                    }
                )
            ],
        )
        result = runner.invoke(
            create,
            cli_args,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )
        assert result.exit_code == 0
        assert "--> Bucket test-bucket has been registered." in result.output


def test_list_buckets():
    runner = CliRunner()
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/customer-resources/buckets",
            json={
                "response": "ok",
                "data": [BUCKET_INFO],
            },
            headers={"Content-Type": "application/json"},
        )
        result = runner.invoke(
            list,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )
        assert result.exit_code == 0
        assert f"{BUCKET_INFO}" in result.output


def test_update_bucket():
    secret_id = str(uuid.uuid4())
    runner = CliRunner()
    cli_args = [
        "--bucket",
        "test-bucket",
        "--storage-type",
        "s3",
        "--secret-id",
        secret_id,
    ]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{UNITTEST_HOST}/api/v1/customer-resources/buckets/test-bucket?storage_type=S3",
            json={
                "response": "ok",
                "data": BUCKET_INFO,
            },
            headers={"Content-Type": "application/json"},
        )

        resp.add(
            resp.PATCH,
            f"{UNITTEST_HOST}/api/v1/customer-resources/buckets/{BUCKET_INFO['id']}",
            json={
                "response": "ok",
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "secret_id": secret_id,
                    }
                )
            ],
        )

        result = runner.invoke(
            update_secret,
            cli_args,
            env={"GANTRY_API_KEY": "test", "GANTRY_LOGS_LOCATION": UNITTEST_HOST},
        )
        assert result.exit_code == 0
        assert f"--> Bucket test-bucket has been updated with key {secret_id}." in result.output



import concurrent.futures
import datetime
import functools
import json
import pprint
import sys
from collections import OrderedDict
from typing import Any, List, NamedTuple, Optional, Set, Union

import pandas as pd
from cachetools import TTLCache, cached
from pandas import DataFrame
from typeguard import typechecked

from gantry.api_client import APIClient
from gantry.exceptions import GantryException, QueryError
from gantry.query.core.constants import MAX_SIZE_CACHE_B, MAX_TTL_CACHE_SECS
from gantry.query.core.queryframe import GantryQueryFrame, QueryInfo
from gantry.query.core.utils import (
    get_application_node_id,
    get_application_views,
    get_last_application_version,
    get_start_end_time_from_view,
    runs_on,
)

MAX_WORKERS = 32  # Arbitrary workers for the 'describe' fetching op in dataframe


@cached(TTLCache(maxsize=MAX_SIZE_CACHE_B, ttl=MAX_TTL_CACHE_SECS))
def _raw_query_cached_request(
    api_client: APIClient, dumped_json_data: str, resource: str = "aggregate"
):
    return api_client.request(
        "POST",
        f"/api/v1/{resource}/query",
        json=json.loads(dumped_json_data),
        raise_for_status=True,
    )


class GantryDataFrame(GantryQueryFrame):
    """
    Two-dimensional dict that supports querying Gantry data.
    """

    def __init__(
        self,
        api_client: APIClient,
        application: str,
        start_time: Union[str, datetime.datetime],
        end_time: Union[str, datetime.datetime],
        version: Optional[Union[int, str]] = None,
        env: Optional[str] = None,
        filters: Optional[List[dict]] = None,
        tags: Optional[dict] = None,
    ):
        """
        :meta private: This is to skip constructor in the docs
        """
        self.start_time, self.end_time = GantryQueryFrame.resolve_time_strings(start_time, end_time)
        application_node_id = get_application_node_id(
            api_client, application, self.start_time, self.end_time, version
        )
        super().__init__(
            api_client,
            QueryInfo(
                application_node_id, application, version, env, self.start_time, self.end_time
            ),
        )
        self.filters = filters or []  # Holds raw dict filters to be used when fetching data
        self.tags = tags  # Holds tags for filtering when fetching data
        self.version = version
        self.env = env
        self.series = self._build_series()

    def create_view(self, name: str) -> None:
        """Create a view from a Gantry Dataframe"""
        # Inline import to avoid circular dependency.
        from gantry.query.main import create_view

        create_view(
            name=name,
            application=self.query_info.application,
            start_time=self.query_info.start_time,
            end_time=self.query_info.end_time,
            version=self.query_info.version,
            tag_filters=self.tags,
            data_filters=self.filters,
        )

    @classmethod
    def from_view(
        cls,
        api_client: APIClient,
        application: str,
        view: str,
        version: Optional[Union[str, int]],
        environment: Optional[str],
        tags: Optional[dict] = None,
    ):
        version = version or get_last_application_version(
            api_client=api_client, application=application
        )
        views = get_application_views(
            api_client=api_client, application=application, version=version, environment=environment
        )
        try:
            env_condition = lambda v: (environment is None) or (  # noqa
                v["tag_filters"].get("env") == environment
            )
            view_obj = [v for v in views if v["name"] == view and env_condition(v)][0]
        except IndexError:
            raise RuntimeError(f"View {view} not found")

        start_time, end_time = get_start_end_time_from_view(view_obj)

        return cls(
            api_client=api_client,
            application=application,
            start_time=start_time,
            end_time=end_time,
            version=version,
            env=view_obj["tag_filters"]["env"],
            filters=view_obj["data_filters"],
            tags=tags,
        )

    def _clone_with_extra_filters(self, filters: "Filters") -> "GantryDataFrame":
        """Private method to clone a dataframe and add additional filters

        This is used when:

            df2 = df1[series > 20]

        We cannot change the state of df1 (ENG-2653), so we generate
        a copy with the new filters
        """
        # Fail in case the provided filters was applied to another df
        for f in filters.filter_objs:
            if f.series.parent_dataframe is not self:
                raise ValueError(
                    "Filter was performed on a different dataframe that "
                    "the one trying to filter. This is not supported."
                )

        # Important to keep existing filters to support
        #    df2 = df1[df1["some-col"] > 20]
        #    df3 = df2[df2["some_col"] < 50]
        return GantryDataFrame(
            api_client=self.api_client,
            application=self.query_info.application,
            version=self.query_info.version,
            env=self.query_info.environment,
            start_time=self.query_info.start_time,
            end_time=self.query_info.end_time,
            filters=[*self.filters, *[f.filter_ for f in filters.filter_objs]],
            tags=self.tags,
        )

    def __getitem__(self, key: str) -> Optional[Any]:
        try:
            return self.__getattr__(key)
        except AttributeError:
            raise KeyError(f"{key} is not a valid column")

    def __getattr__(self, name: str) -> Optional[Any]:
        if isinstance(name, Filters):
            filters = name  # Rename for clarity
            return self._clone_with_extra_filters(filters)

        if name in self.series:
            return self.series.get(name)
        elif name == "dtypes":
            return self._dtypes
        elif hasattr(GantrySeries, name):
            return self._aggregation_factory(name)

        raise AttributeError(f"{name} is not a valid attribute")

    def __len__(self):
        return int(self._aggregation_factory("count")().max(axis=1).values[0])

    def __repr__(self) -> str:
        info = {
            "application": self.query_info.application,
            "from": repr(self.query_info.start_time),
            "to": repr(self.query_info.end_time),
            "series": list(self.series.keys()),
        }
        if (sys.version_info.major, sys.version_info.minor) <= (3, 7):
            return pprint.pformat(info)  # kwarg sort_dicts not supported
        else:
            return pprint.pformat(info, sort_dicts=False)

    @property
    def dtypes(self) -> pd.Series:
        """Return the column types"""
        return pd.Series(self._dtypes).sort_index(ascending=True)

    def head(self, n: int = 5) -> pd.DataFrame:  # type: ignore
        """Fetch the top n entries. Results will include an additional column, `join_id`,
        whose values can be used to log and join feedback events to your Gantry application.

        Args:
            n (int, defaults to 5): Number of entries to fetch

        Returns:
            A pandas dataframe object.
        """
        return self._head(n=n, query_filters=self.filters, tags=self.tags)

    def tail(self, n: int = 5) -> pd.DataFrame:  # type: ignore
        """Fetch the last n entries. Results will include an additional column, `join_id`,
        whose values can be used to log and join feedback events to your Gantry application.

        Args:
            n (int, defaults to 5): Number of entries to fetch

        Returns:
            A pandas dataframe object.
        """
        return self._tail(n=n, query_filters=self.filters, tags=self.tags)

    def fetch(self) -> pd.DataFrame:  # type: ignore
        """Fetch all entries. Results will include an additional column, `join_id`,
        whose values can be used to log and join feedback events to your Gantry application.

        Returns:
            A pandas dataframe object.
        """
        return self._fetch(query_filters=self.filters, tags=self.tags)

    def _aggregation_factory(self, aggregration: str):
        """Generates methods that calculate aggregations on the series
        that compose this dataframe."""

        def apply_aggr(*args, **kw):
            """Will call & display self.series.aggregation() for each series"""
            result = {}
            is_df_output = False
            for k, s in self.series.items():
                data = getattr(s, aggregration)()
                result[k] = data
                if isinstance(data, pd.DataFrame):
                    is_df_output = True

            if not is_df_output:
                result = DataFrame([list(result.values())], columns=list(result.keys()))

            return result

        return apply_aggr

    def _build_series(self) -> OrderedDict:
        ret = OrderedDict()

        for schema_type in ["prediction_datanodes", "feedback_datanodes", "projection_datanodes"]:
            for feature in self.metadata.get(schema_type, []):  # type: ignore
                ret[feature["name"]] = GantrySeries(
                    feature["name"],
                    feature["id"],
                    feature["dtype"],
                    feature["event_type"],
                    self,
                )

        return ret

    def describe(self) -> pd.DataFrame:
        """Print basic stats on the dataframe."""
        table = OrderedDict()  # type: ignore
        # Fetch all stats for series concurrently:
        with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
            future_to_series = OrderedDict()
            for name, s in self.series.items():
                future_to_series[executor.submit(s._describe)] = name
                table[name] = None  # hack to keep table, futures can be solved any order
            for future in concurrent.futures.as_completed(future_to_series):
                table[future_to_series[future]] = future.result()

        headers = list(table.keys())

        # Get a set of stats maintaining order
        x = [stat_name for series, stats in table.items() for stat_name, _ in stats.items()]
        all_stats = sorted(set(x), key=x.index)

        data = [
            [self.query_info.application] * len(headers),
            [self.query_info.start_time] * len(headers),
            [self.query_info.end_time] * len(headers),
        ]

        for stat in all_stats:
            row = [table[series].get(stat, "N/A") for series in headers]
            data.append(row)

        return pd.DataFrame(data, index=["application", "from", "to", *all_stats], columns=headers)


class _FilterObj(NamedTuple):
    """Private container of a raw dict filter and the series
    it applies to.
    """

    filter_: dict
    series: "GantrySeries"


class Filters:
    """Wrapper over a list of filter objects. This object will be used
    as return type when doing 'series > 10' with support for later
    indexing on dataframes or series
    """

    def __init__(self, filter_objs: List[_FilterObj]) -> None:
        self.filter_objs: List[_FilterObj] = filter_objs

    @classmethod
    def _from_single_filter(cls, filter_: dict, series: "GantrySeries") -> "Filters":
        return cls([_FilterObj(filter_, series)])

    def __and__(self, other: "Filters") -> "Filters":
        """Support chaining filters from same dataframe.
        Different series from same dataframes can be &-chained

        """
        parent_df = self.filter_objs[0].series.parent_dataframe

        for f in other.filter_objs:
            if f.series.parent_dataframe != parent_df:
                # Fail in case chaining of filters from different
                # dataframes.
                raise ValueError("Cannot &-compare series from different dataframes")

        return Filters([*self.filter_objs, *other.filter_objs])


SeriesFilterType = Optional[Union[float, int, str, bool, datetime.datetime]]
SortableType = Union[float, int, datetime.datetime]


def customseriestypechecker(func):
    @functools.wraps(func)
    def wrapper(self, other):
        if isinstance(other, GantrySeries):
            raise TypeError(
                "Series cannot be compared to each other directly, but multiple "
                "filters can be chained together using &, for example: "
                "gdf[(gdf[feat_1] != x) & (gdf[feat2] != y)]"
            )

        return func(self, other)

    return wrapper


class GantrySeries(GantryQueryFrame):
    """
    Gantry Series object, similar to Pandas Series.

    Operations on Gantry Series return objects that can be used as masks for
    Gantry Series or Gantry Dataframes.

    .. code-block:: python

       # Operations on series return masks
       mask = some_series.contains("apples")
       apples_series = some_series[mask]

       apples_series.fetch()


    """

    def __init__(
        self,
        name: str,
        id: str,
        datatype: str,
        series_type: str,
        parent_dataframe: GantryDataFrame,
    ) -> None:
        """
        :meta private: This is to skip constructor in the docs
        """
        super().__init__(parent_dataframe.api_client, parent_dataframe.query_info)
        self.name = name
        self.id = id
        self.datatype = datatype
        self.series_type = series_type
        self.parent_dataframe = parent_dataframe

    @property
    def filters(self) -> List[dict]:
        return self.parent_dataframe.filters

    def __getitem__(self, key: Any) -> Optional[Any]:
        if isinstance(key, Filters):
            df = self.parent_dataframe._clone_with_extra_filters(key)
            return df[self.name]

        raise ValueError(
            "GantrySeries object only supports filter masks "
            f"(for example 'series[series > 10]'). {key} is not a valid filter"
        )

    def _filter_builder(self, filter_type: str, value: Any) -> Filters:
        new_filter = {"feature_name": self.name, filter_type: value}
        return Filters._from_single_filter(new_filter, self)

    @runs_on("bool")
    def __invert__(self):
        return self._filter_builder("boolean_query", False)

    @customseriestypechecker
    @typechecked
    def __eq__(self, other: SeriesFilterType):  # type: ignore
        if other is None:
            raise ValueError("use is_none instead of == to check if something is None")
        elif self.dtype == "bool":
            return self._filter_builder("boolean_query", other)
        elif self.dtype in ["str", "uuid"]:
            return self._filter_builder("string_query", other)
        elif self.dtype in ["int", "float"]:
            return self._filter_builder("equals", other)
        else:
            raise ValueError(f"Currently cannot compare {self.dtype} to {other}.")

    @customseriestypechecker
    @typechecked
    def __ne__(self, other: SeriesFilterType):  # type: ignore
        if other is None:
            raise ValueError("use is_none instead of != to check if something is None")
        elif self.dtype == "bool":
            return self._filter_builder("boolean_query", not other)
        elif self.dtype in ["str", "uuid"]:
            return self._filter_builder("string_query", "^(?!({})$).*$".format(other))
        elif self.dtype in ["int", "float"]:
            return self._filter_builder("not_number_query", other)
        else:
            raise ValueError(f"Currently cannot compare {self.type} to {other}.")

    @runs_on("int", "float", "datetime")
    @customseriestypechecker
    @typechecked
    def __gt__(self, other: SortableType):
        return self._filter_builder("lower_bound", other)

    @runs_on("int", "float", "datetime")
    @customseriestypechecker
    @typechecked
    def __ge__(self, other: SortableType):
        return self._filter_builder("inclusive_lower_bound", other)

    @runs_on("int", "float", "datetime")
    @customseriestypechecker
    @typechecked
    def __lt__(self, other: SortableType):
        return self._filter_builder("upper_bound", other)

    @runs_on("int", "float", "datetime")
    @customseriestypechecker
    @typechecked
    def __le__(self, other: SortableType):
        return self._filter_builder("inclusive_upper_bound", other)

    def __len__(self):
        return self.count()

    @runs_on("str")
    @typechecked
    def isin(self, other: List[str]):
        """For string Series, whether an entry is in a list of strings.

        Args:
            other (list[str]): the list of strings

        """
        return self._filter_builder("category_query", other)

    @runs_on("str")
    @typechecked
    def contains(self, other: str):
        """For string Series, whether an entry contains a string

        Args:
            other (str): string to compare

        """
        return self._filter_builder("string_query", other)

    def notnull(self):
        """Filters out null values."""
        return self._filter_builder("null_value", False)

    def notna(self):
        """Alias for notnull"""
        return self.notnull()

    def isnull(self):
        """Filters out non null values (ie, null values remain)"""
        return self._filter_builder("null_value", True)

    def isna(self):
        """Alias for isnull"""
        return self.isnull()

    def __repr__(self) -> str:
        info = {
            "application": self.parent_dataframe.query_info.application,
            "from": repr(self.query_info.start_time),
            "to": repr(self.query_info.end_time),
            "name": self.name,
            "type": self.dtype,
            "series_type": self.series_type,
        }
        if (sys.version_info.major, sys.version_info.minor) <= (3, 7):
            return pprint.pformat(info)  # kwarg sort_dicts not supported
        else:
            return pprint.pformat(info, sort_dicts=False)

    def __getattr__(self, name: str):
        if name == "dtype":
            return self._dtype()

    def _get_dtype(self, column_name: str) -> str:
        return self.parent_dataframe._dtypes.get(column_name)

    def head(self, n: int = 5) -> pd.DataFrame:  # type: ignore
        """Fetch the top n entries

        Args:
            n (int, defaults to 5): Number of entries to fetch

        Returns:
            A pandas dataframe object.
        """
        return self._head(
            n=n,
            column=self.name,
            query_filters=self.parent_dataframe.filters,
            tags=self.parent_dataframe.tags,
        )

    def tail(self, n: int = 5) -> pd.DataFrame:  # type: ignore
        """Fetch the last n entries

        Args:
            n (int, defaults to 5): Number of entries to fetch

        Returns:
            A pandas dataframe object.
        """
        return self._tail(
            n=n,
            column=self.name,
            query_filters=self.parent_dataframe.filters,
            tags=self.parent_dataframe.tags,
        )

    def fetch(self) -> pd.DataFrame:  # type: ignore
        """Fetch all entries

        Returns:
            A pandas dataframe object.
        """
        return self._fetch(
            column=self.name,
            query_filters=self.parent_dataframe.filters,
            tags=self.parent_dataframe.tags,
        )

    @runs_on("int", "float")
    def mean(
        self, num_points: int = 1, group_by: Optional[str] = None
    ) -> Union[float, pd.DataFrame]:
        """
        Runs on int and float series only.
        Get the computed average of this GantrySeries, if available

        Args:
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.


        Returns:
            Float with mean if num_points=1 and group_by=None (the default values),
            else a pd.DataFrame.
        """
        try:
            return self._aggregate_query("mean", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine mean")

    @runs_on("int", "float")
    def std(
        self, num_points: int = 1, group_by: Optional[str] = None
    ) -> Union[float, pd.DataFrame]:
        """
        Runs on int and float series only.
        Get the standard deviation for this GantrySeries, if available

        Args:
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.


        Returns:
            Float with std if num_points=1 and group_by=None (the default values),
            else a pd.DataFrame.
        """
        try:
            return self._aggregate_query("stddev", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine standard deviation")

    @runs_on("int", "float")
    def median(self, num_points: int = 1) -> Union[float, pd.DataFrame]:
        """
        Runs on numeric series only.
        Get the median for this GantrySeries, if available

        Args:
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into

        Returns:
            Float with median in case num_points=1 (the default value),
            else a pd.DataFrame.
        """
        # ToDo: add group_by once linear.app/gantry/issue/ML-729/ is resolved
        try:
            quantiles = self.quantile([0.5], num_points=num_points)
            return quantiles[0.5]
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine median")

    def count(
        self, num_points: int = 1, group_by: Optional[str] = None
    ) -> Union[float, pd.DataFrame]:
        """
        Get the number of available data points for this GantrySeries, if available.
        Works on all series types.

        Args:
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.

        Returns:
            Float with count if num_points=1 and group_by=None (the default values),
            else a pd.DataFrame.
        """
        try:
            return self._aggregate_query("total", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine count")

    @runs_on("int", "float", "datetime")
    def min(
        self, num_points: int = 1, group_by: Optional[str] = None
    ) -> Union[float, datetime.datetime, pd.DataFrame]:
        """
        Runs on int, float, and datetime series only.
        Get the minimum for this GantrySeries, if available.

        Args:
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.


        Returns:
            Float with min value if num_points=1 and group_by=None (the default values),
            else a pd.DataFrame.
        """
        try:
            return self._aggregate_query("minimum", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine min")

    @runs_on("int", "float", "datetime")
    def max(
        self, num_points: int = 1, group_by: Optional[str] = None
    ) -> Union[float, datetime.datetime, pd.DataFrame]:
        """
        Runs on int, float, and datetime series only.
        Get the maximum for this GantrySeries, if available.

        Args:
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.


        Returns:
            Float with max value if num_points=1 and group_by=None (the default values),
            else a pd.DataFrame.
        """
        try:
            return self._aggregate_query("maximum", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine max")

    def histogram(self):
        """
        Runs on bool, int, float and str series only.
        Get histogram of categories for this GantrySeries, if available.

        Gets the histogram of percentages from [0, 1] for available data values
        for this GantrySeries.

        Returns:
            Dict[str, float] histogram
        """
        try:
            hist_res = self._aggregate_query("category_percents")
            return hist_res
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine histogram")

    @runs_on("bool", "str")
    def unique(self) -> Set[str]:
        """
        Runs on bool and str series only.
        Get the unique values in this GantrySeries, if available.

        Returns:
            Set[str] List containing all the unique values that occur in this series.
        """
        try:
            hist_res = self._aggregate_query("category_percents")
            return set(hist_res.keys())
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine unique values")

    @runs_on("int", "float")
    def quantile(self, quantile_vals: List[float] = [0.5]):
        """
        Runs on int and float series only.
        Get quantiles for this GantrySeries, if available.

        Args:
            quantile_vals: list of requested quantiles. Float values in the list should be in the
                range [0, 1].
        Returns:
            List[floats] of len(quantiles_vals) where the order of the outputs matches the input
                quantile_vals order.
        """
        try:
            quantile_res = self._aggregate_query("quantiles", quantile_vals=quantile_vals)
            return {float(k): v for k, v in quantile_res.items()}
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine quantiles")

    def pdf(self):
        """
        Get requested probability density function for this GantrySeries, if available.
        """
        try:
            pdf_res = self._aggregate_query("pdf")
            return pd.DataFrame(pdf_res)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine pdf")

    def cdf(self):
        """
        Get requested cumulative density function for this GantrySeries, if available.
        """
        try:
            cdf_res = self._aggregate_query("cdf")
            return pd.DataFrame(cdf_res)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine cdf")

    def _dtype(self):
        return self.datatype

    def _aggregate_query(
        self,
        attribute,
        num_points: int = 1,
        group_by: Optional[str] = None,
        **attribute_kwargs,
    ):
        data = self.query_info.get_base_query_params()
        if num_points <= 0:
            raise ValueError(f"num_points must be an int > 0, but received value of {num_points}")
        data["num_points"] = num_points
        data["queries"] = {
            "query": {
                "query_type": "feature",
                "model_node_id": self.query_info.application_node_id,
                "stat": attribute,
                "stat_kwargs": {attribute: attribute_kwargs},
                "feature": self.name,
                "filters": self.parent_dataframe.filters,
                "tags": self.parent_dataframe.tags,
            }
        }
        if group_by is not None:
            if group_by not in self.parent_dataframe.series:
                raise ValueError(f"Cannot find column {group_by} in dataframe.")
            if num_points != 1:
                raise ValueError(
                    "Metric computations that specify a group_by feature "
                    "cannot also specify num_points."
                )
            group_by_type = self.parent_dataframe[group_by].datatype  # type: ignore
            data["queries"]["query"]["group_by"] = [
                {"feature_name": group_by, "dtype": group_by_type}
            ]
        resource = "aggregate" if num_points == 1 else "time_series"
        try:
            response = _raw_query_cached_request(
                self.api_client,
                json.dumps(data, default=lambda x: x.isoformat()),
                resource=resource,
            )
            data = response["data"]["query"]
            if group_by is not None:
                df_result = []
                for row in data:
                    result = {}
                    _temp = row["group"][0]  # type: ignore
                    result[_temp["feature_name"]] = _temp["value"]  # type: ignore
                    result[attribute] = row[attribute]
                    df_result.append(result)
                return pd.DataFrame(df_result)
            else:
                if num_points == 1:
                    return data[attribute]
                else:
                    return pd.DataFrame(data["points"])
        except GantryException as e:
            raise GantryException("Aggregation query [" + attribute + "] failed: " + str(e))

    def _describe(self) -> dict:
        c = self.count()
        table = {
            "series type": self.series_type,
            "type": self.dtype,
            "count": c,
            "not null count": self[self.notnull()].count(),  # type: ignore
            "null count": self[self.isnull()].count(),  # type: ignore
        }
        if self.dtype in ["int", "float"]:
            q = self.quantile([0.25, 0.5, 0.75])
            table.update(
                {
                    "q25": q[0.25],
                    "q50": q[0.5],
                    "q75": q[0.75],
                    "mean": self.mean(),
                    "min": self.min(),
                    "max": self.max(),
                }
            )
        if self.dtype == "bool":
            table["True rate"] = self[self == True].count() / c  # type: ignore # noqa

        return table

    def describe(self) -> pd.DataFrame:
        """
        Runs on int, float, and bool series only.
        Return basic stats on the series.

        Returns:
            A pandas dataframe with summary information.

        """
        stats = self._describe()
        return pd.DataFrame(
            [
                self.parent_dataframe.query_info.application,
                self.parent_dataframe.query_info.start_time,
                self.parent_dataframe.query_info.end_time,
                *stats.values(),
            ],
            index=["application", "from", "to", *stats.keys()],
            columns=[self.name],
        )

    @runs_on("bool")
    def percent_true(
        self, dropna: bool = False, num_points: int = 1, group_by: Optional[str] = None
    ):
        """
        Runs on boolean series only.
        Percent true, optionally drop null values.

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.

        Returns:
            Float with true percentage if num_points=1 and group_by=None (by default),
                else pd.DataFrame.
        """
        _temp = self
        if dropna:
            _temp = _temp[_temp.notna()]  # type: ignore
        try:
            return _temp._aggregate_query("percent_true", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine percent_true")

    @runs_on("bool")
    def percent_false(
        self, dropna: bool = False, num_points: int = 1, group_by: Optional[str] = None
    ):
        """
        Runs on boolean series only.
        Percent false, optionally drop null values.

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.

        Returns:
            Float with false percentage if num_points=1 and group_by=None (by default),
                else pd.DataFrame.
        """
        _temp = self
        if dropna:
            _temp = _temp[_temp.notna()]  # type: ignore
        try:
            return _temp._aggregate_query("percent_false", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine percent_false")

    @runs_on("bool")
    def percent_null(self, num_points: int = 1, group_by: Optional[str] = None):
        """
        Runs on boolean series only.
        Percent null/NaN.

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, drop rows with NaN
                values in result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.

        Returns:
            Float with null percentage if num_points=1 and group_by=None (by default),
                else pd.DataFrame.
        """
        try:
            return self._aggregate_query("percent_null", num_points=num_points, group_by=group_by)
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine percent_null.")

    @runs_on("bool")
    def percent_true_not_null(
        self, dropna: bool = False, num_points: int = 1, group_by: Optional[str] = None
    ):
        """
        Runs on boolean series only.
        Percent true of not null values, optionally drop null values before calculating result.

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.

        Returns:
            Float with true percentage if num_points=1 and group_by=None (by default),
                else pd.DataFrame.
        """
        _temp = self
        if dropna:
            _temp = _temp[_temp.notna()]  # type: ignore
        try:
            return _temp._aggregate_query(
                "percent_true_not_null", num_points=num_points, group_by=group_by
            )
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine percent_true_not_null")

    @runs_on("bool")
    def percent_false_not_null(
        self, dropna: bool = False, num_points: int = 1, group_by: Optional[str] = None
    ):
        """
        Runs on boolean series only.
        Percent false of not null values, optionally drop null values before calculating result.

        Args:
            data_node (GantrySeries): GantrySeries which will be calculated
            dropna (bool, defaults to False): if True, first drops NaN values
                before calculating result.
            num_points (int, defaults to 1): number of points to divide the
                time window of the GantrySeries into
            group_by (str, defaults to None): column to use as group_by. If None
                then no group_by operation is performed and a single result is returned.

        Returns:
            Float with false percentage if num_points=1 and group_by=None (by default),
                else pd.DataFrame.
        """
        _temp = self
        if dropna:
            _temp = _temp[_temp.notna()]  # type: ignore
        try:
            return _temp._aggregate_query(
                "percent_false_not_null", num_points=num_points, group_by=group_by
            )
        except (AttributeError, ValueError, KeyError):
            raise QueryError("Invalid query. GantrySeries cannot determine percent_false_not_null")



import mock
import pytest

from gantry.query.client import GantryQuery
from gantry.query.distance import main


@pytest.mark.parametrize(
    "method",
    [
        "d1",
        "dinf",
        "ks",
        "kl",
    ],
)
@mock.patch("gantry.query.main.globals._Query", spec=GantryQuery)
def test_proxy_methods(mock_query, method, series_obj_1, series_obj_2):
    mock_query.distance = mock.Mock()
    getattr(mock_query.distance, method).return_value = "foo"
    assert getattr(main, method)(series_obj_1, series_obj_2) == "foo"
    getattr(mock_query.distance, method).assert_called_once_with(series_obj_1, series_obj_2)



import datetime

import pytest
from pathlib import Path
from freezegun import freeze_time
from gantry.utils import list_all_files

from gantry.utils import _is_offset_naive, check_event_time_in_future

CURRENT_TIME = datetime.datetime(2022, 5, 23, 0, 0, 0)
CLOSE_CURRENT_TIME = datetime.datetime(2022, 5, 23, 1, 0, 0)
FUTURE_TIME = datetime.datetime(2042, 5, 23, 0, 0, 0)
SOME_TIME = datetime.datetime(2022, 5, 10, 17, 50, 32)
ANOTHER_TIME = datetime.datetime(2022, 5, 11, 17, 50, 32)

SOME_TIMEDELTA = datetime.timedelta(hours=-5)
SOME_TIMEZONE = datetime.timezone(SOME_TIMEDELTA)


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize(
    "timestamp",
    [
        FUTURE_TIME,
        FUTURE_TIME.astimezone(datetime.timezone.utc),
        CLOSE_CURRENT_TIME,
        CLOSE_CURRENT_TIME.astimezone(SOME_TIMEZONE),
    ],
)
def test_check_event_time_in_future(timestamp):
    """All passed timestamps will be assumed UTC"""
    assert check_event_time_in_future(timestamp) is True


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize(
    "timestamp", [SOME_TIME, ANOTHER_TIME, ANOTHER_TIME.astimezone(SOME_TIMEZONE)]
)
def test_check_event_time_in_past(timestamp):
    """All passed timestamps will be assumed UTC"""
    assert check_event_time_in_future(timestamp) is False


@pytest.mark.parametrize(
    ["timestamp", "expected"],
    [
        (CURRENT_TIME, True),
        (FUTURE_TIME, True),
        (CURRENT_TIME.astimezone(datetime.timezone.utc), False),
        (FUTURE_TIME.astimezone(), False),
        (ANOTHER_TIME.astimezone(SOME_TIMEZONE), False),
    ],
)
def test_is_offset_naive(timestamp, expected):
    assert _is_offset_naive(timestamp) is expected


@pytest.mark.parametrize(
    ["include_hidden_file", "expected_size"],
    [(True, 4), (False, 1)],
)
def test_list_all_files(include_hidden_file, expected_size, datadir):
    assert len(list_all_files(Path(datadir), include_hidden_file)) == expected_size
    if not include_hidden_file:
        assert list_all_files(Path(datadir), include_hidden_file)[0].name == "visible_test_file.txt"



import mock
import pytest

from gantry.query.client import GantryQuery
from gantry.query.metric import main


@pytest.mark.parametrize(
    ["method", "params"],
    [
        (
            "accuracy_score",
            {"outputs": mock.Mock(), "feedback": mock.Mock(), "dropna": True, "num_points": 1},
        ),
        (
            "mean_squared_error",
            {
                "outputs": mock.Mock(),
                "feedback": mock.Mock(),
                "dropna": True,
                "num_points": 1,
                "multioutput": "uniform_average",
                "squared": True,
            },
        ),
        (
            "confusion_matrix",
            {"outputs": mock.Mock(), "feedback": mock.Mock(), "dropna": True, "num_points": 1},
        ),
        (
            "f1_score",
            {
                "outputs": mock.Mock(),
                "feedback": mock.Mock(),
                "dropna": True,
                "num_points": 1,
                "average": "micro",
            },
        ),
        (
            "r2_score",
            {
                "outputs": mock.Mock(),
                "feedback": mock.Mock(),
                "dropna": True,
                "num_points": 1,
                "multioutput": "uniform_average",
            },
        ),
        (
            "precision_score",
            {
                "outputs": mock.Mock(),
                "feedback": mock.Mock(),
                "dropna": True,
                "num_points": 1,
                "average": "micro",
            },
        ),
        (
            "recall_score",
            {
                "outputs": mock.Mock(),
                "feedback": mock.Mock(),
                "dropna": True,
                "num_points": 1,
                "average": "micro",
            },
        ),
        (
            "roc_auc_score",
            {"outputs": mock.Mock(), "feedback": mock.Mock(), "dropna": True, "num_points": 1},
        ),
        (
            "percent_null",
            {"data_node": mock.Mock(), "dropna": True, "num_points": 1},
        ),
        (
            "percent_true",
            {"data_node": mock.Mock(), "dropna": True, "num_points": 1},
        ),
        (
            "percent_false",
            {"data_node": mock.Mock(), "dropna": True, "num_points": 1},
        ),
    ],
)
@mock.patch("gantry.query.main.globals._Query", spec=GantryQuery)
def test_proxy_methods(mock_query, method, params):
    mock_query.metric = mock.Mock()
    getattr(mock_query.metric, method).return_value = "foo"
    assert getattr(main, method)(**params) == "foo"
    getattr(mock_query.metric, method).assert_called_once_with(**params)



import os
from typing import Optional

from gantry.alerts.main import _init as alerts_init
from gantry.applications.main import _init as applications_init
from gantry.applications.core import Application, Run
from gantry.applications.llm import CompletionApplication, VersionDetails
from gantry.applications.main import (
    archive_application,
    create_application,
    delete_application,
    get_application,
)
from gantry.automations.curators.main import _init as curators_init
from gantry.automations.main import _init as automations_init
from gantry.dataset.main import _init as dataset_init
from gantry.logger.main import LOGGING_LEVEL_T
from gantry.logger.main import _init as logger_init
from gantry.logger.main import (
    get_client,
    log,
    log_file,
    log_from_data_connector,
    log_record,
    log_records,
    ping,
    ready,
    setup_logger,
)
from gantry.logger.utils import JoinKey
from gantry.query.main import _init as query_init
from gantry.utils import compute_feedback_id
from gantry.version import __version__ as _version

__all__ = [
    "init",
    "Application",
    "CompletionApplication",
    "VersionDetails",
    "Run",
    "log",
    "log_file",
    "log_from_data_connector",
    "log_record",
    "log_records",
    "ping",
    "ready",
    "get_client",
    "setup_logger",
    "compute_feedback_id",
    "JoinKey",
    "_version",
    "create_application",
    "get_application",
    "delete_application",
    "archive_application",
]


def init(
    api_key: Optional[str] = None,
    logging_level: Optional[LOGGING_LEVEL_T] = None,
    environment: Optional[str] = None,
    send_in_background: Optional[bool] = None,
):
    """
    Initialize gantry services. Initialization should happen before any interaction with Gantry
    (whether is ingesting data or performing queries).

    After initialization, the SDK logging module, query module and alerts module
    will be enabled.

    Example:

    .. code-block:: python

       import gantry
       gantry.init(api_key="foobar")


    Args:
        api_key (optional, str): The Gantry API Key. You can also set this parameter by setting
            the env variable GANTRY_API_KEY.
        logging_level (optional, str): Set logging level for Gantry system logging.
            You can also set this parameter by setting the env variable GANTRY_LOGGING_LEVEL.
            Options are: DEBUG, INFO, WARNING, CRITICAL or ERROR.
            If not specified, it defaults to INFO.
        environment (optional, str): Set the value for the environment label attached
            to data instrumented. You can also set this parameter by setting the env
            variable GANTRY_ENVIRONMENT. If not provided, it defaults to "dev".
            The environment is essentially a tag attached to data. To override
            this value later, you can set an 'env' tag in the tags parameters
            in the ingestion methods.
        send_in_background (optional, bool): set whether Gantry logging methods should
            run synchronously or not. You can also set this value by setting
            the env variable GANTRY_SEND_IN_BACKGROUND.
            If not provided, it defaults to True unless running in an AWS lambda.

    """
    api_key = _get_api_key(api_key)
    logger_init(api_key, logging_level, environment, send_in_background)
    query_init(api_key)
    alerts_init(api_key)
    curators_init(api_key)
    dataset_init(api_key)
    applications_init(api_key)
    automations_init(api_key)


def _get_api_key(api_key: Optional[str] = None) -> str:
    """
    Get API key from environment if not already supplied. Raise exception if not found.
    """
    api_key = api_key or os.environ.get("GANTRY_API_KEY")
    if api_key is None:
        raise ValueError(
            "Please provide an API key to initialize Gantry SDK. API key can be provided"
            + "as parameter to init methods or by setting env var GANTRY_API_KEY"
        )
    return api_key



import json
import tempfile
from contextlib import nullcontext as does_not_raise
from io import StringIO

import mock
import pandas as pd
import pytest
import responses
from freezegun import freeze_time

from gantry.exceptions import (  # noqa
    GantryBatchCreationException,
    GantryLoggingException,
)
from gantry.logger import utils
from gantry.logger.constants import BatchType, UploadFileType
from gantry.logger.stores import APILogStore
from gantry.logger.types import DataLink, DataLinkElement
from gantry.logger.utils import (
    JoinKey,
    _batch_success_msg,
    _build_batch_iterator,
    _build_success_msg,
    _concurrent_upload_multipart_batch,
    _is_empty,
    _is_not_empty,
    _log_exception,
    _put_block,
)


@pytest.mark.parametrize(
    ["test_obj", "expected_result"],
    [
        (pd.DataFrame(), True),
        (pd.DataFrame.from_dict({"A": [200, 201]}), False),
        ([1, 2, 3], False),
        ([], True),
        (None, True),
    ],
)
def test_is_empty(test_obj, expected_result):
    assert _is_empty(test_obj) is expected_result


@pytest.mark.parametrize(
    ["test_obj", "expected_result"],
    [
        (pd.DataFrame(), False),
        (pd.DataFrame.from_dict({"A": [200, 201]}), True),
        ([1, 2, 3], True),
        ([], False),
        (None, False),
    ],
)
def test_is_not_empty(test_obj, expected_result):
    assert _is_not_empty(test_obj) is expected_result


def test_log_exception():
    try:
        _log_exception(lambda: exec("raise(GantryLoggingException())"))()
    except Exception as e:
        pytest.fail(f"_log_exception() failed due to raising {e}")
    with pytest.raises(GantryBatchCreationException):
        _log_exception(lambda: exec("raise(GantryBatchCreationException())"))()
    try:
        _log_exception(lambda: exec("raise(Exception())"))()
    except Exception as e:
        pytest.fail(f"_log_exception() failed due to raising {e}")


def test_build_batch_iterator():
    event_list = [{1: 1}, {2: 2}, {3: 3}, {4: 4}, {5: 5}]
    # We should only have 3 batches created from the list.
    batch_iter = _build_batch_iterator(event_list, 2)
    iters = 0
    for _ in batch_iter:
        iters += 1
    assert iters == 3

    # We should still have all 5 lines in the file
    batch_iter = _build_batch_iterator(event_list, 2)
    result = "".join([part.decode("utf-8") for part in batch_iter])
    file = StringIO(result)

    for line in file.readlines():
        json.loads(line)

    file.seek(0)
    assert len(file.readlines()) == 5


def test_put_block():
    def custom_matcher(req):
        return (
            req.body == b"foobar",
            "Body doesn't match",
        )

    with responses.RequestsMock() as resp:
        resp.add(
            resp.PUT,
            url="http://foo/bar",
            headers={"ETag": "0123456789", "Content-Type": "application/json"},
            match=[custom_matcher],
        )
        res = _put_block("http://foo/bar", 10, b"foobar")
        assert res == {"ETag": "0123456789", "PartNumber": 10}


@mock.patch("gantry.logger.utils._put_block")
def test_concurrent_upload(mock_put_block):
    mock_put_block.side_effect = [
        {"ETag": "123", "PartNumber": 2},
        {"ETag": "456", "PartNumber": 1},
        {"ETag": "789", "PartNumber": 3},
    ]

    data = [{"1": "foo"}, {"2": "bar"}, {"3": "baz"}]
    signed_urls = ["http://foo", "http://bar", "http://baz"]

    ret = _concurrent_upload_multipart_batch(iter(data), signed_urls)
    assert ret == [
        {"ETag": "456", "PartNumber": 1},
        {"ETag": "123", "PartNumber": 2},
        {"ETag": "789", "PartNumber": 3},
    ]


@pytest.mark.parametrize(
    ["host", "application", "expected"],
    [
        (
            "https://staging.com/",
            "app",
            "Track your batch at https://staging.com/applications/app/jobs",
        ),
        (
            "https://staging.com",
            "app",
            "Track your batch at https://staging.com/applications/app/jobs",
        ),
        (
            "https://staging.com",
            "my app",
            "Track your batch at https://staging.com/applications/my%20app/jobs",
        ),
    ],
)
def test_build_success_msg(host, application, expected):
    assert _build_success_msg(host, application) == expected


@mock.patch("gantry.logger.utils._build_success_msg")
def test_batch_success_msg(mock_build_msg):
    log_store = mock.Mock(__class__=APILogStore)
    _batch_success_msg("12345", "app", log_store)
    mock_build_msg.assert_called_with(log_store._api_client._host, "app")


@pytest.mark.parametrize(
    ["dict_", "expected"],
    [
        ({"foo": "bar", "bar": "baz"}, "44258089824ec9db73fd592f418fbea1"),
        ({"bar": "baz", "foo": "bar"}, "44258089824ec9db73fd592f418fbea1"),
        ({"foo": "bar", "bar": 10}, "01debdea5ba523b44164d202fbbe42da"),
        ({"bar": None, "foo": "bar"}, "ade3188c2cfeae4234e0b7b042b74c35"),
        ({"foo": True, "bar": 10.4321}, "294ca280f5485b49fec1247e5b45d644"),
        ({"bar": [1, 2, 3], "foo": {"another": "dict"}}, "152874e4c1db6061d5a45ee38c3eace0"),
    ],
)
def test_join_key_from_dict(dict_, expected):
    assert JoinKey.from_dict(dict_) == expected


@pytest.mark.parametrize("dict_", [None, True, 10, "something"])
def test_join_key_from_dict_error(dict_):
    with pytest.raises(TypeError):
        _ = JoinKey.from_dict(dict_)


@pytest.mark.parametrize(
    "content, sep, expected",
    [
        (
            """a, b, c
            10, 20, 30""",
            ",",
            ["a", "b", "c"],
        ),
        (
            """a,        b,       c
            10, 20, 30""",
            ",",
            ["a", "b", "c"],
        ),
        (
            """a\tb\tc
            10\t20\t30""",
            "\t",
            ["a", "b", "c"],
        ),
        (
            '''"a",        "b",       "c"
            "foo", "bar", "baz"''',
            ",",
            ["a", "b", "c"],
        ),
    ],
)
def test_csv_columns(content, sep, expected):
    with tempfile.NamedTemporaryFile("w") as tmp:
        tmp.write(content)
        tmp.flush()
        assert utils._get_csv_columns(tmp.name, sep) == expected


@pytest.mark.parametrize(
    "target, columns, expectation",
    [
        ("foo", ["foo", "bar"], does_not_raise()),
        ("foo", ["bar", "baz"], pytest.raises(ValueError)),
    ],
)
def test_validate_column(target, columns, expectation):
    with expectation:
        utils._validate_column(target, columns)


@pytest.mark.parametrize(
    "targets, columns, expectation",
    [
        (["foo"], ["foo", "bar"], does_not_raise()),
        (["foo", "bar"], ["foo", "bar"], does_not_raise()),
        (["foo", "bar"], ["bar", "baz"], pytest.raises(ValueError)),
        (["foo"], ["bar", "baz"], pytest.raises(ValueError)),
    ],
)
def test_validate_columns(targets, columns, expectation):
    with expectation:
        utils._validate_columns(targets, columns)


@pytest.mark.parametrize(
    "inputs, outputs, feedback, expected",
    [
        ({"foo": "bar"}, {"bar": "baz"}, None, BatchType.PREDICTION),
        ({"foo": "bar"}, {"bar": "baz"}, {"baz": "bax"}, BatchType.RECORD),
        ({}, {}, {"baz": "bax"}, BatchType.FEEDBACK),
    ],
)
def test_feature_to_column_mapping_get_batch_type(inputs, outputs, feedback, expected):
    assert (
        utils._FeatureToColumnMapping(
            inputs=inputs,
            outputs=outputs,
            feedback=feedback,
        ).get_batch_type()
        == expected
    )


@pytest.mark.parametrize(
    "inputs, outputs, feedback, feedback_id, expectation",
    [
        ({}, {"foo": "bar"}, {}, {}, pytest.raises(ValueError)),
        ({"bar": "baz"}, {"foo": "bar"}, {}, {}, does_not_raise()),
        ({"bar": "baz"}, {"foo": "bar"}, {"baz": "bax"}, {"foo": "bar"}, does_not_raise()),
        ({"bar": "baz"}, {"foo": "bar"}, {"baz": "bax"}, {}, pytest.raises(ValueError)),
        (None, None, None, {}, pytest.raises(ValueError)),
        ({}, {}, {}, {}, pytest.raises(ValueError)),
    ],
)
def test_feature_to_column_mapping_validate(inputs, outputs, feedback, feedback_id, expectation):
    with expectation:
        utils._FeatureToColumnMapping(
            inputs=inputs, outputs=outputs, feedback=feedback, feedback_id=feedback_id
        ).validate()


@pytest.mark.parametrize(
    "columns, inputs, outputs, feedback, tags, timestamp, feedback_id,"
    "feedback_keys, join_key, expected",
    [
        (
            ["a", "b", "c", "f", "f.id", "foo", "bar"],
            ["a", "b"],
            ["c"],
            ["f"],
            None,
            None,
            "f.id",
            ["foo", "bar"],
            None,
            utils._FeatureToColumnMapping(
                timestamp={},
                inputs={"a": DataLinkElement(ref=0), "b": DataLinkElement(ref=1)},
                outputs={"c": DataLinkElement(ref=2)},
                feedback={"f": DataLinkElement(ref=3)},
                tags={},
                feedback_id={"f.id": DataLinkElement(ref=4)},
                feedback_keys=["foo", "bar"],
            ),
        ),
        (
            ["a", "b", "output_c", "timestamp", "tags.A"],
            ["a", "b"],
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            utils._FeatureToColumnMapping(
                timestamp={"timestamp": DataLinkElement(ref=3)},
                inputs={"a": DataLinkElement(ref=0), "b": DataLinkElement(ref=1)},
                outputs={"output_c": DataLinkElement(ref=2)},
                feedback={},
                tags={"tags.A": DataLinkElement(ref=4)},
                feedback_id={},
                feedback_keys=[],
            ),
        ),
        (
            ["inputs.a", "inputs.b", "c", "other_timestamp", "some_tag"],
            None,
            ["c"],
            None,
            ["some_tag"],
            "other_timestamp",
            None,
            None,
            None,
            utils._FeatureToColumnMapping(
                timestamp={"other_timestamp": DataLinkElement(ref=3)},
                inputs={"inputs.a": DataLinkElement(ref=0), "inputs.b": DataLinkElement(ref=1)},
                outputs={"c": DataLinkElement(ref=2)},
                feedback={},
                tags={"some_tag": DataLinkElement(ref=4)},
                feedback_id={},
                feedback_keys=[],
            ),
        ),
        (
            ["a", "b", "c", "f", "join_key", "foo", "bar"],
            ["a", "b"],
            ["c"],
            ["f"],
            None,
            None,
            None,
            None,
            None,
            utils._FeatureToColumnMapping(
                timestamp={},
                inputs={"a": DataLinkElement(ref=0), "b": DataLinkElement(ref=1)},
                outputs={"c": DataLinkElement(ref=2)},
                feedback={"f": DataLinkElement(ref=3)},
                tags={},
                feedback_id={"join_key": DataLinkElement(ref=4)},
                feedback_keys=[],
                _join_key={"join_key": DataLinkElement(ref=4)},
            ),
        ),
        (
            ["a", "b", "c", "f", "other_join_key", "foo", "bar"],
            ["a", "b"],
            ["c"],
            ["f"],
            None,
            None,
            None,
            None,
            "other_join_key",
            utils._FeatureToColumnMapping(
                timestamp={},
                inputs={"a": DataLinkElement(ref=0), "b": DataLinkElement(ref=1)},
                outputs={"c": DataLinkElement(ref=2)},
                feedback={"f": DataLinkElement(ref=3)},
                tags={},
                feedback_id={"other_join_key": DataLinkElement(ref=4)},
                feedback_keys=[],
                _join_key={"other_join_key": DataLinkElement(ref=4)},
            ),
        ),
    ],
)
@mock.patch("gantry.logger.utils._validate_column")
@mock.patch("gantry.logger.utils._validate_columns")
def test_feature_to_column_mapping_from_columns(
    mock_validate_columns,
    mock_validate_column,
    columns,
    inputs,
    outputs,
    feedback,
    tags,
    timestamp,
    feedback_id,
    feedback_keys,
    join_key,
    expected,
):
    assert (
        utils._FeatureToColumnMapping.from_columns(
            columns,
            inputs,
            outputs,
            feedback,
            tags,
            timestamp,
            feedback_id,
            feedback_keys,
            join_key,
        )
        == expected
    )
    mock_validate_columns.assert_called()
    if timestamp or feedback_id:
        mock_validate_column.assert_called()
    else:
        mock_validate_column.assert_not_called()


@pytest.mark.parametrize(
    "columns, inputs, outputs, feedback, tags, timestamp, feedback_id," "feedback_keys, join_key",
    [
        (
            ["a", "b", "c", "f", "other_join_key", "foo", "bar", "feedback_id"],
            ["a", "b"],
            ["c"],
            ["f"],
            None,
            None,
            None,
            None,
            "other_join_key",
        ),
        (
            ["a", "b", "c", "f", "other_join_key", "foo", "bar", "feedback_id"],
            ["a", "b"],
            ["c"],
            ["f"],
            None,
            None,
            "feedback_id",
            None,
            "other_join_key",
        ),
        (
            ["a", "b", "c", "f", "other_join_key", "foo", "bar", "feedback_id"],
            ["a", "b"],
            ["c"],
            ["f"],
            None,
            None,
            "feedback_id",
            ["a", "b"],
            "other_join_key",
        ),
    ],
)
@mock.patch("gantry.logger.utils._validate_column")
@mock.patch("gantry.logger.utils._validate_columns")
def test_feature_to_column_mapping_from_columns_deprecation_error(
    mock_validate_columns,
    mock_validate_column,
    columns,
    inputs,
    outputs,
    feedback,
    tags,
    timestamp,
    feedback_id,
    feedback_keys,
    join_key,
):
    with pytest.raises(ValueError):
        utils._FeatureToColumnMapping.from_columns(
            columns,
            inputs,
            outputs,
            feedback,
            tags,
            timestamp,
            feedback_id,
            feedback_keys,
            join_key,
        )


@freeze_time("2012-01-14")
@mock.patch("gantry.logger.utils.get_file_linecount")
@mock.patch("gantry.logger.utils._get_csv_columns")
@mock.patch("gantry.logger.utils._FeatureToColumnMapping.from_columns")
def test_build_data_link_from_file_deprecated(
    mock_from_cols, mock_get_csv_columns, mock_get_file_linecount
):
    mock_get_file_linecount.return_value = 1001
    mock_get_csv_columns.return_value = ["foo", "bar"]
    validate_mock = mock.Mock()
    mock_from_cols.return_value = mock.Mock(
        inputs="parsed_inputs",
        outputs="parsed_outputs",
        feedback="parsed_feedback",
        timestamp="parsed_ts",
        feedback_id="parsed_fid",
        feedback_keys="parsed_fkeys",
        tags="parsed_tags",
        get_batch_type=mock.Mock(return_value=BatchType.PREDICTION),
        validate=validate_mock,
    )

    assert utils._build_data_link_from_file(
        application="my-app",
        filepath="/some/file/to/upload",
        sep="\t",
        version="1.2.3",
        timestamp="some-timestamp",
        inputs=["a", "b"],
        outputs=["c"],
        feedback=["f"],
        tags=["foo", "bar"],
        feedback_id="baz",
        feedback_keys=["c"],
        join_key=None,
    ) == DataLink(
        file_type=UploadFileType.CSV_WITH_HEADERS,
        batch_type=BatchType.PREDICTION,
        num_events=1000,
        application="my-app",
        version="1.2.3",
        log_timestamp="2012-01-14T00:00:00",
        inputs="parsed_inputs",
        outputs="parsed_outputs",
        feedback="parsed_feedback",
        timestamp="parsed_ts",
        feedback_id="parsed_fid",
        feedback_keys="parsed_fkeys",
        tags="parsed_tags",
    )

    mock_get_csv_columns.assert_called_with("/some/file/to/upload", "\t")
    mock_from_cols.assert_called_with(
        ["foo", "bar"],
        ["a", "b"],
        ["c"],
        ["f"],
        ["foo", "bar"],
        "some-timestamp",
        "baz",
        ["c"],
        None,
    )
    validate_mock.assert_called_once()


@freeze_time("2012-01-14")
@mock.patch("gantry.logger.utils.get_file_linecount")
@mock.patch("gantry.logger.utils._get_csv_columns")
@mock.patch("gantry.logger.utils._FeatureToColumnMapping.from_columns")
def test_build_data_link_from_file(mock_from_cols, mock_get_csv_columns, mock_get_file_linecount):
    mock_get_file_linecount.return_value = 1001
    mock_get_csv_columns.return_value = ["foo", "bar"]
    validate_mock = mock.Mock()
    mock_from_cols.return_value = mock.Mock(
        inputs="parsed_inputs",
        outputs="parsed_outputs",
        feedback="parsed_feedback",
        timestamp="parsed_ts",
        feedback_id="some_join_key",
        feedback_keys=None,
        tags="parsed_tags",
        get_batch_type=mock.Mock(return_value=BatchType.PREDICTION),
        validate=validate_mock,
    )

    assert utils._build_data_link_from_file(
        application="my-app",
        filepath="/some/file/to/upload",
        sep="\t",
        version="1.2.3",
        timestamp="some-timestamp",
        inputs=["a", "b"],
        outputs=["c"],
        feedback=["f"],
        tags=["foo", "bar"],
        join_key="some_join_key",
        feedback_id=None,
        feedback_keys=None,
    ) == DataLink(
        file_type=UploadFileType.CSV_WITH_HEADERS,
        batch_type=BatchType.PREDICTION,
        num_events=1000,
        application="my-app",
        version="1.2.3",
        log_timestamp="2012-01-14T00:00:00",
        inputs="parsed_inputs",
        outputs="parsed_outputs",
        feedback="parsed_feedback",
        timestamp="parsed_ts",
        feedback_id="some_join_key",
        feedback_keys=None,
        tags="parsed_tags",
    )

    mock_get_csv_columns.assert_called_with("/some/file/to/upload", "\t")
    mock_from_cols.assert_called_with(
        ["foo", "bar"],
        ["a", "b"],
        ["c"],
        ["f"],
        ["foo", "bar"],
        "some-timestamp",
        None,
        None,
        "some_join_key",
    )
    validate_mock.assert_called_once()


def test_build_data_link_integration_deprecated(datadir):
    csv_file = datadir / "feedback_deprecated.csv"
    data_link = utils._build_data_link_from_file(
        application="test_app",
        filepath=str(csv_file),
        sep=",",
        version="1.2.3",
        feedback_id=None,
        feedback=["test_float", "test_int"],
        timestamp=None,
        inputs=None,
        outputs=None,
        tags=None,
        feedback_keys=None,
        join_key=None,
    )

    assert data_link.application == "test_app"
    assert data_link.version == "1.2.3"
    assert data_link.file_type == UploadFileType.CSV_WITH_HEADERS
    assert data_link.batch_type == BatchType.FEEDBACK
    assert data_link.feedback_id["feedback_id"].ref == 0
    assert data_link.feedback["test_float"].ref == 1
    assert data_link.feedback["test_int"].ref == 2


def test_build_data_link_integration(datadir):
    csv_file = datadir / "feedback.csv"
    data_link = utils._build_data_link_from_file(
        application="test_app",
        filepath=str(csv_file),
        sep=",",
        version="1.2.3",
        feedback_id=None,
        feedback=["test_float", "test_int"],
        timestamp=None,
        inputs=None,
        outputs=None,
        tags=None,
        feedback_keys=None,
        join_key="some_join_key",
    )

    assert data_link.application == "test_app"
    assert data_link.version == "1.2.3"
    assert data_link.file_type == UploadFileType.CSV_WITH_HEADERS
    assert data_link.batch_type == BatchType.FEEDBACK
    assert data_link.feedback_id["some_join_key"].ref == 0
    assert data_link.feedback["test_float"].ref == 1
    assert data_link.feedback["test_int"].ref == 2



import datetime

import mock
import pytest
import responses

from gantry.api_client import APIClient
from gantry.query.client import GantryQuery
from gantry.query.core import dataframe, queryframe
from gantry.query.core.dataframe import GantryDataFrame, GantrySeries

ORIGIN = "https://gantry-api:12345"
START_TIME = datetime.datetime(2008, 9, 3, 20, 56, 35)
END_TIME = datetime.datetime(2008, 9, 10, 23, 58, 23)
SOME_DURATION = datetime.timedelta(days=7, hours=1, minutes=45, seconds=30)

START_TIME_STR = START_TIME.isoformat()
END_TIME_STR = END_TIME.isoformat()

DATA = {
    "id": "ABCD1234",
    "prediction_datanodes": [
        {
            "name": "pred_datanode",
            "dtype": "intA",
            "event_type": "A",
            "node_metadata": {
                "content_retrieval": ["inputs", "pred_datanode"],
                "dtype": {"id_str": False, "is_categorical": False, "long_str": False},
            },
            "id": "ABCD1234",
        },
    ],
    "feedback_datanodes": [
        {
            "name": "feed_datanode",
            "dtype": "intB",
            "event_type": "B",
            "node_metadata": {
                "content_retrieval": ["inputs", "feed_datanode"],
                "dtype": {"id_str": False, "is_categorical": False, "long_str": False},
            },
            "id": "ABCD1234",
        },
    ],
    "projection_datanodes": [
        {
            "name": "proj_datanode",
            "dtype": "intC",
            "event_type": "C",
            "node_metadata": {
                "content_retrieval": ["inputs", "proj_datanode"],
                "dtype": {"id_str": False, "is_categorical": False, "long_str": False},
            },
            "id": "ABCD1234",
        },
    ],
}


MULTIPLE_FEATURES_INT_DATA = {
    "id": "ABCD1234",
    "prediction_datanodes": [
        {
            "name": "A",
            "dtype": "int",
            "event_type": "prediction",
            "node_metadata": {
                "content_retrieval": ["inputs", "pred_datanode"],
                "dtype": {"id_str": False, "is_categorical": False, "long_str": False},
            },
            "id": "ABCD1234",
        },
        {
            "name": "B",
            "dtype": "int",
            "event_type": "prediction",
            "node_metadata": {
                "content_retrieval": ["inputs", "pred_datanode"],
                "dtype": {"id_str": False, "is_categorical": False, "long_str": False},
            },
            "id": "ABCD1234",
        },
    ],
    "feedback_datanodes": [],
    "projection_datanodes": [],
}


def to_tz_time(dt_timestamp):
    return dt_timestamp.isoformat() + "Z"


@pytest.fixture
def query_info_obj():
    application_node_id = "af0d0317-adbf-4524-b838-9dddf1c43fc8"
    application = "foobar"
    version = "1.2.3"
    environment = "dev"

    return queryframe.QueryInfo(
        application_node_id, application, version, environment, START_TIME, END_TIME
    )


@pytest.fixture
def api_client_obj():
    return APIClient(origin=ORIGIN, api_key="abcd1234")


@pytest.fixture
def gantry_query_obj(api_client_obj):
    return GantryQuery(api_client_obj)


@pytest.fixture
@mock.patch("gantry.query.core.dataframe.GantryDataFrame._build_series")
def df_factory(mock_populate, api_client_obj):
    def wrapper(filters, tags=None):
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                "{}/api/v1/models/foobar/schemas".format(ORIGIN),
                json={"response": "ok", "data": DATA},
            )

            return GantryDataFrame(
                api_client=api_client_obj,
                application="foobar",
                version="1.2.3",
                env="dev",
                start_time=START_TIME,
                end_time=END_TIME,
                filters=filters,
                tags=tags,
            )

    return wrapper


@pytest.fixture
def series_obj_1(api_client_obj, query_info_obj):
    series = GantrySeries(
        name="foo",
        datatype="int",
        series_type="prediction",
        parent_dataframe=mock.Mock(api_client=api_client_obj, query_info=query_info_obj),
        id="abc",
    )
    series.count = mock.Mock(return_value=10)
    return series


@pytest.fixture
def series_obj_2(api_client_obj, query_info_obj):
    series = GantrySeries(
        name="bar",
        datatype="int",
        series_type="prediction",
        parent_dataframe=mock.Mock(api_client=api_client_obj, query_info=query_info_obj),
        id="abc",
    )
    series.count = mock.Mock(return_value=10)
    return series


@pytest.fixture
def series_obj(api_client_obj, query_info_obj):
    return dataframe.GantrySeries(
        name="series_name",
        datatype="int",
        series_type="prediction",
        parent_dataframe=mock.Mock(api_client=api_client_obj, query_info=query_info_obj),
        id="abc123",
    )


@pytest.fixture
def series_obj_str(api_client_obj, query_info_obj):
    return dataframe.GantrySeries(
        name="series_name",
        datatype="str",
        series_type="prediction",
        parent_dataframe=mock.Mock(api_client=api_client_obj, query_info=query_info_obj),
        id="abc123",
    )


@pytest.fixture
def series_obj_bool(api_client_obj, query_info_obj):
    return dataframe.GantrySeries(
        name="A",
        datatype="bool",
        series_type="prediction",
        parent_dataframe=mock.Mock(api_client=api_client_obj, query_info=query_info_obj),
        id="abc123",
    )


@pytest.fixture
def series_obj_factory(api_client_obj, query_info_obj):
    def wrapper(datatype, parent_dataframe=None, name="series_name", count=10):
        series = dataframe.GantrySeries(
            name=name,
            datatype=datatype,
            series_type="prediction",
            parent_dataframe=parent_dataframe
            if parent_dataframe is not None
            else mock.Mock(api_client=api_client_obj, query_info=query_info_obj),
            id="abc123",
        )
        series.count = mock.Mock(return_value=count)
        return series

    return wrapper


@pytest.fixture
def series_obj_factory_with_filters_and_tags(df_factory, api_client_obj, query_info_obj):
    def wrapper(datatype, filters, tags):
        return dataframe.GantrySeries(
            name="series_name",
            datatype=datatype,
            series_type="prediction",
            parent_dataframe=df_factory(filters=filters, tags=tags),
            id="abc123",
        )

    return wrapper



from datasets import Sequence, Value

GANTRY_FOLDER = ".dataset_metadata"
DATASET_MANIFEST_FILE = f"{GANTRY_FOLDER}/.gantry_manifest.jsonl"  # manifest file
DATASET_HEAD_FILE = f"{GANTRY_FOLDER}/HEAD"  # current head commit info
STASH_FOLDER = f"{GANTRY_FOLDER}/.dataset_stash"
DATASET_STASH_FILE = f"{STASH_FOLDER}/.stash.json"
HF_FOLDER = f"{GANTRY_FOLDER}/huggingface"
TABULAR_MANIFESTS = "tabular_manifests"
ARTIFACTS = "artifacts"
DATASET_CONFIG_FILE = "dataset_config.yaml"
DATASET_FEATURES_KEY = "features"
DATASET_FEEDBACK_KEY = "labels"
BACKUP_SUFFIX = "_backup"
NEW_SUFFIX = "_new"
CSV_SUFFIX = ".csv"

FILE_NAME = "file_name"
NEW_FILES = "new_files"
MODIFIED_FILES = "modified_files"
DELETED_FILES = "deleted_files"
UNCHANGED_FILES = "unchanged_files"
SHA256 = "sha256"
URL = "url"


METADATA_S3_FILE_VERSION = "metadata_s3_file_version"


FILE_PATH = "file_path"
OBJ_KEY = "obj_key"
VERSION_ID = "version_id"

# TODO: add support to cast image/video/audio from string to the file
GANTRY_2_HF_DTYPE = {
    "Float": Value(dtype="float64", id=None),
    "Text": Value(dtype="string", id=None),
    "Integer": Value(dtype="int64", id=None),
    "Boolean": Value(dtype="bool", id=None),
    "Categorical": Value(dtype="string", id=None),
    "UUID": Value(dtype="string", id=None),
    "ID": Value(dtype="string", id=None),
    "Datetime": Value(dtype="timestamp[ns, tz=UTC]", id=None),
    "Unix_Time": Value(dtype="int64", id=None),
    "Json": dict(),
    "Image": Value(dtype="string", id=None),
    "Audio": Value(dtype="string", id=None),
    "Video": Value(dtype="string", id=None),
    "File": Value(dtype="string", id=None),
    "Chat_Message": Value(dtype="string", id=None),
    "Array<LLM_Generation_Choice>": Value(dtype="string", id=None),
    "Array<Chat_Message>": Value(dtype="string", id=None),
    "Array<String>": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
    "Array<Float>": Sequence(feature=Value(dtype="float64", id=None), length=-1, id=None),
    "Array<Integer>": Sequence(feature=Value(dtype="int64", id=None), length=-1, id=None),
    "Array<Boolean>": Sequence(feature=Value(dtype="bool", id=None), length=-1, id=None),
    "Array<UUID>": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
    "Array<ID>": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
    "Unknown": Value(dtype="string", id=None),
}

EMPTY_STR_SHA256 = "47DEQpj8HBSa+/TImW+5JCeuQeRkm5NMpJWZG3hSuFU="



"""
Exposes top-level aliases for using the alerts client.
"""

import os
from typing import Optional
from urllib.parse import urlparse

from gantry.alerts import globals
from gantry.alerts.client import GantryAlerts
from gantry.alerts.globals import _alerts_alias, validate_init
from gantry.api_client import APIClient
from gantry.const import PROD_API_URL


def _init(api_key: Optional[str] = None):
    backend = os.environ.get("GANTRY_LOGS_LOCATION", PROD_API_URL)
    parsed_origin = urlparse(backend)
    if parsed_origin.scheme not in ("http", "https"):
        raise ValueError(
            "Invalid backend. http or https backends " + "supported. Got {}".format(backend)
        )

    api_key = os.environ.get("GANTRY_API_KEY") if api_key is None else api_key

    if not api_key:
        raise ValueError(
            """
            No API key provided. Please pass the api_key parameter or set the GANTRY_API_KEY
            environment variable.
            """
        )

    api_client = APIClient(backend, api_key)
    globals._Alerts = GantryAlerts(api_client)  # type: ignore[union-attr]


@_alerts_alias
def get_alerts(*args, **kwargs):
    validate_init()
    return globals._Alerts.get_alerts(*args, **kwargs)  # type: ignore[union-attr]


@_alerts_alias
def create_alert(*args, **kwargs):
    validate_init()
    return globals._Alerts.create_alert(*args, **kwargs)  # type: ignore[union-attr, return-value]


@_alerts_alias
def update_alert(*args, **kwargs):
    validate_init()
    return globals._Alerts.update_alert(*args, **kwargs)  # type: ignore[union-attr, return-value]


@_alerts_alias
def create_slack_notification(*args, **kwargs):
    validate_init()
    return globals._Alerts.create_slack_notification(  # type: ignore[union-attr, return-value]
        *args, **kwargs
    )


@_alerts_alias
def get_slack_notifications(*args, **kwargs):
    validate_init()
    return globals._Alerts.get_slack_notifications(*args, **kwargs)  # type: ignore[union-attr]


@_alerts_alias
def delete_alert(*args, **kwargs):
    validate_init()
    return globals._Alerts.delete_alert(*args, **kwargs)  # type: ignore[union-attr, return-value]


@_alerts_alias
def delete_slack_notification(*args, **kwargs):
    validate_init()
    return globals._Alerts.delete_slack_notification(  # type: ignore[union-attr, return-value]
        *args, **kwargs
    )



import logging

import click

from gantry import _version
from gantry.cli.application import application
from gantry.cli.bucket import bucket
from gantry.cli.data_connector import data_connector
from gantry.cli.projection import projection
from gantry.cli.secret import secret
from gantry.cli.labeling import labeling


@click.group()
@click.option(
    "--loglevel",
    "-l",
    type=click.Choice(["debug", "info", "warning", "error", "critical"], case_sensitive=False),
    default="info",
)
@click.version_option(_version)
@click.pass_context
def cli(ctx, loglevel):
    ctx.ensure_object(dict)

    formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    pkg_logger = logging.getLogger("gantry")
    pkg_logger.addHandler(handler)
    pkg_logger.setLevel(level=loglevel.upper())

    # pass logging info for subcommands to configure other loggers
    ctx.obj["logging_handler"] = handler
    ctx.obj["logging_level"] = loglevel.upper()


cli.add_command(projection)
cli.add_command(application)
cli.add_command(bucket)
cli.add_command(data_connector)
cli.add_command(secret)
cli.add_command(labeling)

if __name__ == "__main__":
    cli()
